/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gausselim;

import gausselim.pivoting.PivotManager;
import gausselim.systems.ExampleManager;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.math.BigDecimal;
import java.math.MathContext;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Main window for the Gaussian Elimination applet.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class GaussElimFrame extends JFrame implements ChangeListener {
	
	private static final int DEFAULT_ORDER = 3;
	private static final int DEFAULT_PRECISION = 16;
	private static final int DEFAULT_FIELD_WIDTH = 6;
	
	private GaussElimModel gem;
	// we need this so we can check for order changes
	private int previousOrder;

	private int fieldWidth;
	// whether or not we should reuse the last used input
	private boolean saveValues;
	// last used input
	private BigDecimal[][] customA;
	private BigDecimal[] customB;
	
	// the matrix panels use a CardLayout (one stack for A and L, one for b and x)
	private JPanel alPanel, aPanel, lPanel, bxPanel, bPanel, xPanel;
	private JFormattedTextField[][] aFields, lFields;
	private JFormattedTextField[] bFields, xFields;
	private JRadioButton aButton, bButton;	
	private GaussElimSettings ges;
	private JButton stepButton, solveButton;
	private JLabel statusLabel;
	
	private ResourceBundle bundle = ResourceBundle.getBundle("properties/gausselim", Locale.getDefault());
	
	/**
	 * Create the main window.
	 */
	public GaussElimFrame() {
		super(ResourceBundle.getBundle("properties/gausselim", Locale.getDefault()).getString("mainWindowTitle"));
		
		gem = new GaussElimModel(DEFAULT_ORDER, new MathContext(DEFAULT_PRECISION));
		gem.addChangeListener(this);
		// setting this to zero makes sure the panels get built on startup
		previousOrder = 0;
		
		fieldWidth = DEFAULT_FIELD_WIDTH;
		saveValues = true;		
		
		// build interface
		setup();
		setResizable(false);
		
		// update interface
		stateChanged(new ChangeEvent(this));
		ges.stateChanged(new ChangeEvent(this));
	}
	
	// build interface
	private void setup() {
		alPanel = new JPanel(new CardLayout());
		bxPanel = new JPanel(new CardLayout());
		
		add(alPanel, BorderLayout.CENTER);		
		add(bxPanel, BorderLayout.EAST);
		
		ActionListener viewListener = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// move the correct cards on top
				updateView();
			}
		};
		
		// since A and L are stacked together, they cannot be viewed at the same time (same goes for b and x)
		ButtonGroup alGroup = new ButtonGroup();
		ButtonGroup bxGroup = new ButtonGroup();
		
		aButton = new JRadioButton(bundle.getString("buttonA"), true);		
		bButton = new JRadioButton(bundle.getString("buttonB"), true);
		JRadioButton lButton = new JRadioButton(bundle.getString("buttonL"), false);
		JRadioButton xButton = new JRadioButton(bundle.getString("buttonX"), false);
		
		aButton.addActionListener(viewListener);
		lButton.addActionListener(viewListener);
		bButton.addActionListener(viewListener);
		xButton.addActionListener(viewListener);
		
		alGroup.add(aButton);		
		alGroup.add(lButton);
		bxGroup.add(bButton);
		bxGroup.add(xButton);
		
		JPanel choicesPanel = new JPanel(new GridLayout(1, 4, 5, 0));
		choicesPanel.add(aButton);
		choicesPanel.add(lButton);
		choicesPanel.add(bButton);
		choicesPanel.add(xButton);

		JButton settingsButton = new JButton(bundle.getString("buttonSettings"));
		JButton resetButton = new JButton(bundle.getString("buttonReset"));
		stepButton = new JButton(bundle.getString("buttonStep"));
		solveButton = new JButton(bundle.getString("buttonSolve"));
		
		ges = new GaussElimSettings();
		settingsButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// show or hide the settings window
				ges.setVisible(! ges.isVisible());
			}
		});

		resetButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				gem.reset();
				// optionally restore the last used input (if not null)
				if (saveValues && customA != null && customB != null)
					gem.initialize(customA, customB);
			}
		});
		
		stepButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// save and write back the user's input
				if (gem.getStep() == 1) {
					saveCustomValues();
					gem.initialize(customA, customB);
				}
				gem.step();
			}
		});
		
		solveButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// finish all steps first
				// save and write back the user's input
				if (gem.getStep() == 1) {
					saveCustomValues();
					gem.initialize(customA, customB);
				}
				// compare current step to previous step to make sure the model isn't stuck in a loop (if it cannot perform a particular step)
				int previousStep = 0;
				while (gem.getStep() < gem.getOrder() && previousStep < gem.getStep()) {
					previousStep = gem.getStep();
					gem.step();
				}
				if (gem.getStep() == gem.getOrder())
					gem.solve();
			}
		});
		
		JPanel toolbar = new JPanel(new GridLayout(1, 4, 5, 0));
		toolbar.add(settingsButton);
		toolbar.add(resetButton);
		toolbar.add(stepButton);
		toolbar.add(solveButton);
		
		statusLabel = new JLabel("");
		JPanel statusPanel = new JPanel(new BorderLayout());
		statusPanel.setBorder(BorderFactory.createEtchedBorder());
		statusPanel.add(statusLabel, BorderLayout.SOUTH);

		JPanel utilPanel = new JPanel(new GridLayout(3, 1, 0, 5));
		utilPanel.add(choicesPanel);
		utilPanel.add(toolbar);
		utilPanel.add(statusPanel);
		
		add(utilPanel, BorderLayout.SOUTH);
	}
	
	// swap panels
	private void updateView() {
		
		if (aButton.isSelected())
			((CardLayout)alPanel.getLayout()).show(alPanel, "A");
		else
			((CardLayout)alPanel.getLayout()).show(alPanel, "L");
		
		if (bButton.isSelected())
			((CardLayout)bxPanel.getLayout()).show(bxPanel, "b");
		else
			((CardLayout)bxPanel.getLayout()).show(bxPanel, "x");
	}

	// save last used input
	private void saveCustomValues() {
		int order = gem.getOrder();
		customA = new BigDecimal[order][order];
		customB = new BigDecimal[order];
		
		for (int i = 0; i < order; i++) {
			for (int j = 0; j < order; j++)
				customA[i][j] = (BigDecimal) aFields[i][j].getValue();
			customB[i] = (BigDecimal) bFields[i].getValue();
		}
	}
	
	/**
	 * Sets the width of the fields.
	 */
	public void setFieldWidth(int fieldWidth) {
		this.fieldWidth = fieldWidth;
		int order = gem.getOrder();
		
		for (int i = 0; i < order; i++) {
			for (int j = 0; j < order; j++) {
				aFields[i][j].setColumns(fieldWidth);
				lFields[i][j].setColumns(fieldWidth);
			}
			bFields[i].setColumns(fieldWidth);
			xFields[i].setColumns(fieldWidth);
		}
		
		// the size of the cards changed, so we update the size of the stacks as well
		alPanel.setPreferredSize(aPanel.getPreferredSize());
		bxPanel.setPreferredSize(bPanel.getPreferredSize());
		
		pack();
	}
	
	/**
	 * Enable/disable the reuse of last used input.
	 */
	public void setSaveValues(boolean saveVales) {
		this.saveValues = saveVales;
	}
	
	/**
	 * Update the main window.
	 */
	public void stateChanged(ChangeEvent e) {
		int order = gem.getOrder();
		
		// check if the order changed and if so rebuild the panels
		if (order != previousOrder) {
			previousOrder = order;
			// if the order changed, we forget the last used input
			customA = null;
			customB = null;
			rebuildPanels();
		}

		// update matrix elements
		readValues();
		
		// only allow editing at step 1
		for (int i = 0; i < order; i++) {
			for (int j = 0; j < order; j++)
				aFields[i][j].setEditable(gem.getStep() == 1);
			bFields[i].setEditable(gem.getStep() == 1);
		}
		
		// update panel borders
		aPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "A <" + gem.getStep() + ">"));
		lPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "L <" + gem.getStep() + ">"));
		bPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "b <" + gem.getStep() + ">"));
		xPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "x"));
		
		// update toolbar status
		stepButton.setEnabled(gem.getStep() < order);
		solveButton.setEnabled(! gem.isSolved());
		statusLabel.setText(gem.getStatus());
	}

	// build new panels
	private void rebuildPanels() {
		int order = gem.getOrder();
		
		aPanel = new JPanel(new GridLayout(order, order, 5, 5));
		lPanel = new JPanel(new GridLayout(order, order, 5, 5));	
		bPanel = new JPanel(new GridLayout(order, 1, 0, 5));
		xPanel = new JPanel(new GridLayout(order, 1, 0, 5));
		
		// we need to add borders here as well so the panels don't grow after their size is calculated/used
		aPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "A <1>"));
		lPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "L <1>"));
		bPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "b <1>"));
		xPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "x"));
		
		aFields = new JFormattedTextField[order][order];
		lFields = new JFormattedTextField[order][order];
		bFields = new JFormattedTextField[order];
		xFields = new JFormattedTextField[order];
		
		for (int i = 0; i < order; i++) {
			for (int j = 0; j < order; j++) {
				aFields[i][j] = new JFormattedTextField(new RealFractionFormatter(gem.getMathContext()));
				lFields[i][j] = new JFormattedTextField(new RealFractionFormatter(gem.getMathContext()));
				aFields[i][j].setColumns(fieldWidth);
				lFields[i][j].setColumns(fieldWidth);
				lFields[i][j].setEditable(false);
				aPanel.add(aFields[i][j]);
				lPanel.add(lFields[i][j]);
			}
			bFields[i] = new JFormattedTextField(new RealFractionFormatter(gem.getMathContext()));
			xFields[i] = new JFormattedTextField(new RealFractionFormatter(gem.getMathContext()));
			bFields[i].setColumns(fieldWidth);
			xFields[i].setColumns(fieldWidth);
			xFields[i].setEditable(false);
			bPanel.add(bFields[i]);
			xPanel.add(xFields[i]);
		}
		
		// replace previous panels
		alPanel.add(aPanel, "A");
		alPanel.add(lPanel, "L");
		bxPanel.add(bPanel, "b");
		bxPanel.add(xPanel, "x");
		
		// show the correct cards
		if (aButton.isSelected())
			((CardLayout)alPanel.getLayout()).show(alPanel, "A");
		else
			((CardLayout)alPanel.getLayout()).show(alPanel, "L");
		
		if (bButton.isSelected())
			((CardLayout)bxPanel.getLayout()).show(bxPanel, "b");
		else
			((CardLayout)bxPanel.getLayout()).show(bxPanel, "x");
		
		// the size of the cards changed, so we update the size of the stacks as well
		alPanel.setPreferredSize(aPanel.getPreferredSize());
		bxPanel.setPreferredSize(bPanel.getPreferredSize());
		
		pack();
	}

	// update all field values
	private void readValues() {
		int order = gem.getOrder();
		
		for (int i = 1; i <= order; i++) {
			for (int j = 1; j <= order; j++) {
				aFields[i - 1][j - 1].setValue(gem.getA(i, j));
				lFields[i - 1][j - 1].setValue(gem.getL(i, j));
			}
			bFields[i - 1].setValue(gem.getB(i));
			xFields[i - 1].setValue(gem.getX(i));
		}
	}
	
	/**
	 * Simple test program.
	 */
	public static void main(String[] args) {
		GaussElimFrame gef = new GaussElimFrame();
		gef.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		gef.setVisible(true);
	}
	
	// Settings window
	private class GaussElimSettings extends JDialog implements ChangeListener {
		
		private ExampleManager em;
		private PivotManager pm;
		
		private JSpinner orderSpinner, precisionSpinner, fieldWidthSpinner;
		private JComboBox examples, pivots;
		private JCheckBox customValues;
		
		public GaussElimSettings() {
			// non-modal dialog
			super(GaussElimFrame.this, bundle.getString("settingsWindowTitle"), false);
			
			em = new ExampleManager();
			pm = new PivotManager();
			
			// build interface
			setup();
			setResizable(false);
		}
		
		// build interface
		private void setup() {
			// labels need to go in a panel so they get some padding
			JPanel orderLabelPanel = new JPanel();
			JPanel precisionLabelPanel = new JPanel();
			JPanel fieldWidthLabelPanel = new JPanel();
			JPanel examplesLabelPanel = new JPanel();
			JPanel pivotsLabelPanel = new JPanel();
			
			orderLabelPanel.add(new JLabel(bundle.getString("labelOrder")));
			precisionLabelPanel.add(new JLabel(bundle.getString("labelPrecision")));
			fieldWidthLabelPanel.add(new JLabel(bundle.getString("labelFieldWidth")));
			examplesLabelPanel.add(new JLabel(bundle.getString("labelExamples")));
			pivotsLabelPanel.add(new JLabel(bundle.getString("labelPivoting")));
			
			orderSpinner = new JSpinner(new SpinnerNumberModel(DEFAULT_ORDER, 2, 16, 1));
			precisionSpinner = new JSpinner(new SpinnerNumberModel(DEFAULT_PRECISION, 3, 16, 1));
			fieldWidthSpinner = new JSpinner(new SpinnerNumberModel(DEFAULT_FIELD_WIDTH, 3, 16, 1));
			
			orderSpinner.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					gem.setOrder((Integer) orderSpinner.getValue());
				}
			});
			
			precisionSpinner.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					gem.setMathContext(new MathContext((Integer) precisionSpinner.getValue()));
				}
			});
			
			fieldWidthSpinner.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					setFieldWidth((Integer) fieldWidthSpinner.getValue());
				}
			});
			
			examples = new JComboBox();
			pivots = new JComboBox();
			
			for (String example : em.getDescriptions())
				examples.addItem(example);
			for (String pivot : pm.getDescriptions())
				pivots.addItem(pivot);
			
			examples.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					// only change to a different value (avoid change/update loops)
					String item = (String) examples.getSelectedItem();
					if (! em.getDescription(gem.getExampleSystem()).equals(item))
						gem.setExampleSystem(em.getExampleSystem(item));					
				}
			});
			
			pivots.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					// only change to a different value (avoid change/update loops)
					String item = (String) pivots.getSelectedItem();
					if (! pm.getDescription(gem.getPivotingMethod()).equals(item)) {
						if (gem.getStep() == 1) {
							// changing pivoting methods causes a ChangeEvent, so we need to save the user's input first
							// we don't use saveCustomValues() since that is for last _used_ input
							int order = gem.getOrder();
							BigDecimal [][] a = new BigDecimal[order][order];
							BigDecimal [] b = new BigDecimal[order];
							
							for (int i = 0; i < order; i++) {
								for (int j = 0; j < order; j++)
									a[i][j] = (BigDecimal) aFields[i][j].getValue();
								b[i] = (BigDecimal) bFields[i].getValue();
							}
							gem.initialize(a, b);
						}
						gem.setPivotingMethod(pm.getPivotingMethod(item));
					}
					
				}
			});
			
			customValues = new JCheckBox(bundle.getString("buttonSaveLastInput"), saveValues);
			customValues.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					setSaveValues(customValues.isSelected());
				}
			});
			
			getContentPane().setLayout(new GridBagLayout());
			GridBagConstraints gbc = new GridBagConstraints();
			
			gbc.gridx = 0;
			gbc.anchor = GridBagConstraints.WEST;
			gbc.insets = new Insets(5, 0, 0, 0);
			
			gbc.gridy = 0;
			add(orderLabelPanel, gbc);
			
			gbc.gridy = 1;
			add(precisionLabelPanel, gbc);
			
			gbc.gridy = 2;
			add(fieldWidthLabelPanel, gbc);
			
			gbc.gridy = 3;
			add(examplesLabelPanel, gbc);
			
			gbc.gridy = 4;
			add(pivotsLabelPanel, gbc);
			
			gbc.gridy = 5;
			gbc.gridwidth = 2;
			add(customValues, gbc);
			
			gbc.gridx = 1;
			gbc.gridwidth = 1;
			gbc.fill = GridBagConstraints.BOTH;
			
			gbc.gridy = 0;
			add(orderSpinner, gbc);
			
			gbc.gridy = 1;
			add(precisionSpinner, gbc);
			
			gbc.gridy = 2;
			add(fieldWidthSpinner, gbc);
			
			gbc.gridy = 3;
			add(examples, gbc);
			
			gbc.gridy = 4;
			add(pivots, gbc);
			
			pack();
		}
		
		/**
		 * Update the settings window.
		 */
		public void stateChanged(ChangeEvent e) {
			orderSpinner.setValue(gem.getOrder());
			precisionSpinner.setValue(gem.getMathContext().getPrecision());
			examples.setSelectedItem(em.getDescription(gem.getExampleSystem()));
			pivots.setSelectedItem(pm.getDescription(gem.getPivotingMethod()));
		}
	}
}