/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gausselim;

import gausselim.pivoting.NoPivoting;
import gausselim.pivoting.PivotingMethod;
import gausselim.systems.ExampleSystem;
import gausselim.systems.ZeroSystem;

import java.math.BigDecimal;
import java.math.MathContext;
import java.text.MessageFormat;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

/**
 * Gaussian elimination for solving systems of linear equations.
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class GaussElimModel {
	
	// we don't keep a separate matrix L but replace the subdiagonal entries of A
	private BigDecimal[][] a;
	private BigDecimal[] b, x;
	
	private int order, step;
	private boolean solved;
	private String status;
	
	// keeps track of the number of significant digits to use
	private MathContext mc;
	// fills input matrices with default values
	private ExampleSystem es;
	// takes care of pivoting
	private PivotingMethod pm;
	
	private EventListenerList ll;
	
	private ResourceBundle bundle = ResourceBundle.getBundle("properties/gausselim", Locale.getDefault());
	
	/**
	 * Create a new model for Gaussian elimination.
	 */
	public GaussElimModel(int order, MathContext mc) {
		// make sure all settings have sensible values
		this.order = (order < 2 ? 2 : order);
		this.mc = (mc == null ? MathContext.DECIMAL64 : mc);		
		es = new ZeroSystem();
		pm = new NoPivoting();
		
		ll = new EventListenerList();
		
		rebuild();
	}

	// build the system
	private void rebuild() {
		a = new BigDecimal[order][order];
		b = new BigDecimal[order];
		x = new BigDecimal[order];
		
		es.fill(a, b, mc);
		for (int i = 0; i < order; i++)
			x[i] = null;
		
		step = 1;
		solved = false;
		status = bundle.getString("statusReady");
	}

	/**
	 * Reset the system.
	 */
	public void reset() {
		rebuild();
		fireStateChanged(new ChangeEvent(this));
	}
	
	/**
	 * Initialize the system with the specified input matrices.
	 * This is only allowed at step 1.
	 */
	public void initialize(BigDecimal[][] a, BigDecimal[] b) {
		if (step == 1) {
			for (int i = 0; i < order; i++) {
				for (int j = 0; j < order; j++)
					this.a[i][j] = a[i][j];
				this.b[i] = b[i];
			}
			fireStateChanged(new ChangeEvent(this));
		}		
	}
	
	/**
	 * @return A(row, column)
	 */
	public BigDecimal getA(int row, int column) {
		// don't return values of L
		if (row > column && column < step)
			return new BigDecimal("0.0", mc);
		else
			return a[row - 1][column - 1];
	}

	/**
	 * @return L(row, column)
	 */
	public BigDecimal getL(int row, int column) {
		// don't return values of A
		if (row == column)
			return new BigDecimal("1.0", mc);
		else if (column > row || column >= step)
			return new BigDecimal("0.0", mc);
		else
			return a[row - 1][column - 1];
	}
	
	/**
	 * Assign a new value to A(row, column).
	 * This is only allowed at step 1.
	 */
	public void setA(int row, int column, BigDecimal value) {
		if (step == 1) {
			a[row - 1][column - 1] = value;
			fireStateChanged(new ChangeEvent(this));
		}
	}

	/**
	 * @return b(row)
	 */
	public BigDecimal getB(int row) {
		return b[row - 1];
	}

	/**
	 * Assign a new value to b(row).
	 * This is only allowed at step 1.
	 */
	public void setB(int row, BigDecimal value) {
		if (step == 1) {
			b[row - 1] = value;
			fireStateChanged(new ChangeEvent(this));
		}		
	}

	/**
	 * @return x(row)
	 */
	public BigDecimal getX(int row) {
		return x[row - 1];
	}

	/**
	 * @return the order of the system
	 */
	public int getOrder() {
		return order;
	}

	/**
	 * Assign a new order (>= 2) to the system.
	 * This will cause the system to reset.
	 */
	public void setOrder(int order) {
		this.order = (order < 2 ? 2 : order);
		reset();
	}

	/**
	 * @return the current step in the elimination process
	 */
	public int getStep() {
		return step;
	}
	
	/**
	 * @return whether the solution has been calculated
	 */
	public boolean isSolved() {
		return solved;
	}
	
	/**
	 * @return the status of the system
	 */
	public String getStatus() {
		return status;
	}
	
	/**
	 * @return the current MathContext being used
	 */
	public MathContext getMathContext() {
		return mc;
	}
	
	/**
	 * Assign a new MathContext to the system.
	 * This will cause the system to reset.
	 * In case of a null value, the default MathContext.DECIMAL64 is used.
	 */
	public void setMathContext(MathContext mc) {
		this.mc = (mc == null ? MathContext.DECIMAL64 : mc);
		reset();
	}
	
	/**
	 * @return the current ExampleSystem being used
	 */
	public ExampleSystem getExampleSystem() {
		return es;
	}
	
	/**
	 * Assign a new ExampleSystem to the system.
	 * This will cause the system to reset.
	 * In case of a null value, the default ZeroSystem is used.
	 */
	public void setExampleSystem(ExampleSystem es) {
		this.es = (es == null ? new ZeroSystem() : es);
		reset();
	}

	/**
	 * @return the current PivotingMethod being used
	 */
	public PivotingMethod getPivotingMethod() {
		return pm;
	}
	
	/**
	 * Assign a new PivotingMethod to the system.
	 * In case of a null value, the default NoPivoting is used.
	 */
	public void setPivotingMethod(PivotingMethod pm) {
		this.pm = (pm == null ? new NoPivoting() : pm);
		fireStateChanged(new ChangeEvent(this));
	}

	/**
	 * Perform a single elimination step.
	 */
	public void step() {
		if (step == order)
			status = bundle.getString("statusFinishedTrySolve");
		else {
			// pivot
			int[] pivot = pm.pivot(a, b, step, mc);
		
			// check if pivoting could find a nonzero pivot element
			if (a[step - 1][step - 1].compareTo(BigDecimal.ZERO) == 0)
				status = MessageFormat.format(bundle.getString("statusCannotPerformStep"), step);
			else {
				for (int i = step; i < order; i++) {
					// L(i,k) = A(i,k)/A(k,k)
					a[i][step - 1] = a[i][step - 1].divide(a[step - 1][step - 1], mc);
					for (int j = step; j < order; j++)
						// A(i,j) = A(i,j) - L(i,k)*A(k,j)
						a[i][j] = a[i][j].subtract(a[i][step - 1].multiply(a[step - 1][j], mc), mc);
					// b(i) = b(i) - L(i,k)*b(k)
					b[i] = b[i].subtract(a[i][step - 1].multiply(b[step - 1], mc), mc);
				}
				
				if (pivot[0] > step || pivot[1] > step)
					status = MessageFormat.format(bundle.getString("statusPerformedStepWithPivot"), step, pivot[0], pivot[1]);
				else
					status = MessageFormat.format(bundle.getString("statusPerformedStep"), step);
				step++;
			}
		}
		fireStateChanged(new ChangeEvent(this));
	}

	/**
	 * Solve the system using backwards substitution.
	 */
	public void solve() {
		if (solved)
			status = bundle.getString("statusAlreadySolved");
		else if (step != order) 
			status = bundle.getString("statusFinishStepsFirst");
		else if (a[order - 1][order - 1].compareTo(BigDecimal.ZERO) == 0)
			status = bundle.getString("statusCannotSolve");
		else {
			for (int i = order - 1; i >= 0; i--) {
				// x(i) = (b(i) - sum(A(i,j)*x(j), j=i+1..order))/A(i,i)
				x[i] = b[i];
				for (int j = i + 1; j < order; j++)
					x[i] = x[i].subtract(a[i][j].multiply(x[j], mc), mc);
				x[i] = x[i].divide(a[i][i], mc);
			}
			solved = true;
			status = bundle.getString("statusSolved");
		}				
		fireStateChanged(new ChangeEvent(this));
	}
	
	/**
	 * Add a ChangeListener.
	 */
	public void addChangeListener(ChangeListener l) {
		ll.add(ChangeListener.class, l);
	}

	/**
	 * Remove a ChangeListener.
	 */
	public void removeChangeListener(ChangeListener l) {
		ll.remove(ChangeListener.class, l);
	}

	/**
	 * Notify all ChangeListeners.
	 */
	public void fireStateChanged(ChangeEvent e) {
		ChangeListener[] listeners = ll.getListeners(ChangeListener.class);
		for (ChangeListener l : listeners)
			l.stateChanged(e);
	}
}