/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gausselim;

import java.math.BigDecimal;
import java.math.MathContext;
import java.text.ParseException;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.text.DefaultFormatter;

/**
 * Formatter for BigDecimals that can parse simple fractions (BigDecimal/BigDecimal).
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RealFractionFormatter extends DefaultFormatter {

	private MathContext mc;
	
	private ResourceBundle bundle = ResourceBundle.getBundle("properties/gausselim", Locale.getDefault());
	
	/**
	 * Create a formatter that parses BigDecimals.
	 */
	public RealFractionFormatter(MathContext mc) {
		this.mc = mc;
	}
	
	/**
	 * Parse the specified string into a BigDecimal.
	 * The first forward slash (if present) is used as the division sign.
	 * The BigDecimal(String, MathContext) constructor is used to parse the numerator and denominator (if present).
	 * In case of a division by zero only the numerator is returned.
	 */
	public Object stringToValue(String string) throws ParseException {
		int sign = string.indexOf('/');

		try {
			if (sign == -1)
				return new BigDecimal(string, mc);
			else {
				BigDecimal numerator, denominator;
			
				numerator = new BigDecimal(string.substring(0, sign), mc);
				denominator = new BigDecimal(string.substring(sign + 1), mc);
				
				if (denominator.compareTo(BigDecimal.ZERO) != 0)
					return numerator.divide(denominator, mc);
				else
					return numerator;
			}
		} catch (NumberFormatException e) {
			throw new ParseException(bundle.getString("Cannot parse this value."), 67);
		}
	}

	/**
	 * Format a BigDecimal.
	 * Uses the toString() method.
	 * A null value is formatted as unknown.
	 */
	public String valueToString(Object value) throws ParseException {
		return value == null ? bundle.getString("rffUnknownValue") : value.toString();
	}
}