/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gausselim.pivoting;

import java.math.BigDecimal;
import java.math.MathContext;

/**
 * Row pivoting (also called partial pivoting).
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class RowPivoting implements PivotingMethod {

	/**
	 * Perform row pivoting.
	 * At step k, this method selects the first row i so that i >= k and |A(i,k)| >= |A(j,k)| for any other j >= k.
	 * Rows i and k are then swapped.
	 */
	public int[] pivot(BigDecimal[][] a, BigDecimal[] b, int step, MathContext mc) {
		int order = a.length;
		
		// Find pivot row
		int pivotRow = step;
		BigDecimal maxValue = a[step - 1][step - 1].abs(mc);
		for (int i = step; i < order; i++)
			if (a[i][step - 1].abs(mc).compareTo(maxValue) > 0) {
				pivotRow = i + 1;
				maxValue = a[i][step - 1].abs(mc);
			}
		
		// Swap rows if necessary
		if (pivotRow > step) {
			for (int i = 0; i < order; i++) {
				BigDecimal temp = a[step - 1][i];
				a[step - 1][i] = a[pivotRow - 1][i];
				a[pivotRow - 1][i] = temp;
			}
			BigDecimal temp = b[step - 1];
			b[step - 1] = b[pivotRow - 1];
			b[pivotRow - 1] = temp;
		}
		
		return new int[] {pivotRow, step};
	}

}
