/**
 * Copyright 2007 Steven Van Impe, Department Of Applied Mathematics And Computer Science, Ghent University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package gausselim.pivoting;

import java.math.BigDecimal;
import java.math.MathContext;

/**
 * Scaled row pivoting (also called implicit pivoting).
 * 
 * @author Steven Van Impe (steven.vanimpe@ugent.be)
 */
public class ScaledRowPivoting implements PivotingMethod {

	/**
	 * Perform scaled row pivoting.
	 * At step k, this method selects the first row i so that i >= k and |A(i,k)|/M(i) >= |A(j,k)|/M(j) for any other j >= k, where M(i) is the largest absolute value in row i.
	 * Rows i and k are then swapped.
	 */
	public int[] pivot(BigDecimal[][] a, BigDecimal[] b, int step, MathContext mc) {
		int order = a.length;
		
		// Find the largest absolute value in each row
		BigDecimal[] maxPerRow = new BigDecimal[order - step + 1];
		for (int i = 0; i < maxPerRow.length; i++) {
			maxPerRow[i] = b[step - 1 + i].abs(mc);
			for (int j = 0; j < maxPerRow.length; j++)
				if (a[step - 1 + i][step - 1 + j].abs(mc).compareTo(maxPerRow[i]) > 0)
					maxPerRow[i] = a[step - 1 + i][step - 1 + j].abs(mc);
		}
		
		// Find pivot row using scaled values
		int pivotRow = step;
		BigDecimal maxValue = a[step - 1][step - 1].divide(maxPerRow[0], mc).abs(mc);
		for (int i = 1; i < maxPerRow.length; i++)
			if (a[step - 1 + i][step - 1].divide(maxPerRow[i], mc).abs(mc).compareTo(maxValue) > 0) {
				pivotRow = step + i;
				maxValue = a[step - 1 + i][step - 1].divide(maxPerRow[i], mc).abs(mc);
			}
		
		// Swap rows if necessary
		if (pivotRow > step) {
			for (int i = 0; i < order; i++) {
				BigDecimal temp = a[step - 1][i];
				a[step - 1][i] = a[pivotRow - 1][i];
				a[pivotRow - 1][i] = temp;
			}
			BigDecimal temp = b[step - 1];
			b[step - 1] = b[pivotRow - 1];
			b[pivotRow - 1] = temp;
		}
		
		return new int[] {pivotRow, step};
	}

}
